import { TxProcessorResult } from "../processor/tx-processor-result.js";
import { Utils } from "../utils/utils.js";
export class BoltzmannResult extends TxProcessorResult {
    constructor(duration, r) {
        super(r.nbCmbn, r.matLnkCombinations, r.matLnkProbabilities, r.entropy, r.dtrmLnksById, r.txos, r.fees, r.intraFees, r.efficiency, r.nbCmbnPrfctCj, r.nbTxosPrfctCj);
        this.dtrmLnks =
            r.dtrmLnksById == null
                ? [[""]]
                : this.replaceDtrmLinks(r.dtrmLnksById, r.txos);
        this.duration = duration;
    }
    replaceDtrmLinks(dtrmLinks, txos) {
        const result = Array.from({
            length: dtrmLinks.size,
        }).fill([]);
        const outs = [...txos.outputs.keys()];
        const ins = [...txos.inputs.keys()];
        let i = 0;
        for (const dtrmLink of dtrmLinks) {
            const out = outs[dtrmLink[0]];
            const n = ins[dtrmLink[1]];
            result[i] = [out, n];
            i++;
        }
        return result;
    }
    toJSON() {
        return {
            nbCmbn: this.nbCmbn,
            matLnkCombinations: this.matLnkCombinations,
            matLnkProbabilities: this.matLnkProbabilities,
            entropy: this.entropy,
            dtrmLnksById: [...this.dtrmLnksById.values()],
            dtrmLnks: this.dtrmLnks,
            txos: {
                inputs: [...this.txos.inputs.entries()],
                outputs: [...this.txos.outputs.entries()],
            },
            fees: this.fees,
            intraFees: {
                feesMaker: this.intraFees.feesMaker,
                feesTaker: this.intraFees.feesTaker,
                hasFees: this.intraFees.hasFees,
            },
            efficiency: this.efficiency,
            nbCmbnPrfctCj: this.nbCmbnPrfctCj?.toString() ?? null,
            nbTxosPrfctCj: this.nbTxosPrfctCj,
        };
    }
    print() {
        console.log(`Inputs = ${JSON.stringify([...this.txos.inputs.entries()])}`);
        console.log(`Outputs = ${JSON.stringify([...this.txos.outputs.entries()])}`);
        console.log(`Fees = ${this.fees} satoshis`);
        if (this.intraFees != null &&
            this.intraFees.feesMaker > 0 &&
            this.intraFees.feesTaker > 0) {
            console.log(`Hypothesis: Max intrafees received by a participant = ${this.intraFees.feesMaker} satoshis`);
            console.log(`Hypothesis: Max intrafees paid by a participant = ${this.intraFees.feesTaker} satoshis`);
        }
        if (this.nbCmbnPrfctCj != null) {
            console.log(`Perfect coinjoin = ${this.nbCmbnPrfctCj} combinations (for ${this.nbTxosPrfctCj.nbIns}x${this.nbTxosPrfctCj.nbOuts})`);
        }
        console.log(`Nb combinations = ${this.nbCmbn}`);
        if (this.entropy != null) {
            console.log(`Tx entropy = ${this.entropy} bits`);
            console.log(`Entropy density = ${this.density}`);
        }
        if (this.efficiency != null) {
            console.log(`Wallet efficiency = ${(this.efficiency ?? 0) * 100}% (${Math.log2(this.efficiency ?? 0)} bits)`);
        }
        if (this.matLnkCombinations == null) {
            if (this.nbCmbn === 0) {
                console.log("Skipped processing of this transaction (too many inputs and/or outputs)");
            }
        }
        else {
            if (this.matLnkProbabilities != null) {
                console.log("Linkability Matrix (probabilities):");
                console.log(JSON.stringify(this.matLnkProbabilities));
            }
            if (this.matLnkCombinations != null) {
                console.log("Linkability Matrix (#combinations with link):");
                console.log(JSON.stringify(this.matLnkCombinations));
            }
        }
        if (this.dtrmLnks == null) {
            console.log("Deterministic links: all");
        }
        else if (this.dtrmLnks.length > 0) {
            console.log("Deterministic links:");
            for (const dtrmLink of this.dtrmLnks) {
                const output = dtrmLink[0];
                const input = dtrmLink[1];
                console.log(`${input} & ${output} are deterministically linked`);
            }
        }
        else {
            console.log("Deterministic links: none");
        }
        console.log("Benchmarks:");
        const benchmarks = [];
        benchmarks.push(["duration", this.duration]);
        console.log(`Duration = ${Utils.duration(this.duration)}`);
        const maxMem = Utils.getMaxMemUsed();
        benchmarks.push(["maxMem", maxMem]);
        console.log(`Max mem used: ${maxMem}M`);
        for (const progress of Utils.getProgressResult()) {
            console.log(progress.getResult());
            benchmarks.push([
                progress.name,
                progress.target,
                progress.computeElapsed() / 1000,
                progress.rate,
                progress.msg,
            ]);
        }
        const exportObj = {};
        exportObj.ins = [...this.txos.inputs.entries()];
        exportObj.outs = [...this.txos.outputs.entries()];
        exportObj.nbCmbn = this.nbCmbn;
        exportObj.mat =
            this.matLnkCombinations == null
                ? null
                : JSON.stringify(this.matLnkCombinations);
        exportObj.benchmarks = benchmarks;
        try {
            const exportStr = JSON.stringify(exportObj);
            console.log(`Export: ${exportStr}`);
        }
        catch (error) {
            console.error(error);
        }
    }
}
//# sourceMappingURL=boltzmann-result.js.map