import { TxosAggregator } from "../aggregator/txos-aggregator.js";
import { LogLevel } from "../beans/boltzmann-settings.js";
import { TxProcessorConst } from "../processor/tx-processor-const.js";
import { PackType, } from "../utils/interfaces.js";
import { ListsUtils } from "../utils/lists-utils.js";
import { CustomMap, Utils } from "../utils/utils.js";
import { TxosLinkerResult } from "./txos-linker-result.js";
export class TxosLinker {
    constructor(fees, maxDuration, maxTxos, logger) {
        this.packs = [];
        this.maxDuration = TxosLinker.MAX_DURATION;
        this.feesOrig = fees;
        this.maxDuration = maxDuration;
        this.maxTxos = maxTxos;
        this.Logger = logger;
    }
    process(txos, linkedTxos, options, intraFees) {
        if (linkedTxos != null && linkedTxos.length > 0) {
            txos = this.packLinkedTxos(linkedTxos, txos);
        }
        if (options.mergeFees && this.feesOrig > 0) {
            this.fees = 0;
            txos.outputs.set(TxosLinker.MARKER_FEES, this.feesOrig);
        }
        else {
            this.fees = this.feesOrig;
        }
        const aggregator = new TxosAggregator(this.Logger);
        const nbOuts = txos.outputs.size;
        const nbIns = txos.inputs.size;
        const hasIntraFees = intraFees?.hasFees ?? false;
        let nbCmbn = 0;
        let matLnk = ListsUtils.newIntMatrix(nbOuts, nbIns, 0);
        let allAgg = this.prepareData(txos);
        txos = { inputs: allAgg.inAgg.txos, outputs: allAgg.outAgg.txos };
        let aggMatches = aggregator.matchAggByVal(allAgg, this.fees, intraFees);
        let dtrmLnks = new Set();
        if (options.precheck && this.checkLimitOk(txos) && !hasIntraFees) {
            if (this.Logger.logLevel === LogLevel.DEBUG) {
                Utils.logMemory(this.Logger, "# PRECHECK");
            }
            dtrmLnks = aggregator.checkDtrmLinks(txos, allAgg, aggMatches);
            if (dtrmLnks.size > 0) {
                for (const dtrmLnk of dtrmLnks) {
                    matLnk[dtrmLnk[0]][dtrmLnk[1]] = 1;
                }
            }
        }
        if (nbIns === 0 || nbOuts === 0) {
            nbCmbn = 1;
            for (const line of matLnk) {
                ListsUtils.fill(line, 1, line.length);
            }
        }
        else if (options.linkability && this.checkLimitOk(txos)) {
            if (this.Logger.logLevel === LogLevel.DEBUG) {
                Utils.logMemory(this.Logger, "# LINKABILITY");
            }
            if (dtrmLnks.size > 0) {
                Utils.logMemory(this.Logger, `PACK·${dtrmLnks.size}·deterministic·links`);
                const txosFinal = txos;
                const dtrmCoordsList = [];
                for (const array of dtrmLnks) {
                    const set = new Set();
                    set.add(String([...txosFinal.outputs.keys()][array[0]]));
                    set.add(String([...txosFinal.inputs.keys()][array[1]]));
                    dtrmCoordsList.push(set);
                }
                txos = this.packLinkedTxos(dtrmCoordsList, txos);
                allAgg = this.prepareData(txos);
                txos = { inputs: allAgg.inAgg.txos, outputs: allAgg.outAgg.txos };
                aggMatches = aggregator.matchAggByVal(allAgg, this.fees, intraFees);
            }
            const matInAggCmbn = aggregator.computeInAggCmbn(aggMatches);
            const result = aggregator.computeLinkMatrix(txos, allAgg, aggMatches, matInAggCmbn, this.maxDuration);
            nbCmbn = result.nbCmbn;
            matLnk = result.matLnkCombinations;
            dtrmLnks = aggregator.findDtrmLinks(matLnk, nbCmbn);
        }
        if (this.packs.length > 0) {
            if (this.Logger.logLevel === LogLevel.DEBUG) {
                Utils.logMemory(this.Logger, `#·UNPACK·${this.packs.length}·packs`);
            }
            const unpackResult = this.unpackLinkMatrix(matLnk, txos);
            txos = unpackResult.txos;
            matLnk = unpackResult.matLnk;
            dtrmLnks = aggregator.findDtrmLinks(matLnk, nbCmbn);
        }
        return new TxosLinkerResult(nbCmbn, matLnk, dtrmLnks, txos);
    }
    packLinkedTxos(linkedTxos, txos) {
        let idx = this.packs.length;
        const packedTxos = {
            inputs: new CustomMap([...txos.inputs.entries()]),
            outputs: new CustomMap([...txos.outputs.entries()]),
        };
        const newPacks = ListsUtils.mergeSets(linkedTxos);
        for (const pack of newPacks) {
            const ins = [];
            let valIns = 0;
            for (const inPack of pack) {
                if (inPack.startsWith(TxProcessorConst.MARKER_INPUT)) {
                    const inPackValue = packedTxos.inputs.get(inPack);
                    ins.push([inPack, inPackValue]);
                    valIns += inPackValue;
                    packedTxos.inputs.delete(inPack);
                }
            }
            idx++;
            if (ins.length > 0) {
                const lbl = `${TxosLinker.MARKER_PACK}${idx}`;
                packedTxos.inputs.set(lbl, valIns);
                this.packs.push({ lbl, packType: PackType.INPUTS, ins, outs: [] });
            }
        }
        return packedTxos;
    }
    unpackLinkMatrix(matLnk, txos) {
        let matRes = [...matLnk];
        let newTxos = {
            inputs: new CustomMap([...txos.inputs.entries()]),
            outputs: new CustomMap([...txos.outputs.entries()]),
        };
        const reversedPacks = [...this.packs].reverse();
        for (let i = this.packs.length - 1; i >= 0; i--) {
            const pack = this.packs[i];
            const result = this.unpackLinkMatrix2(matRes, newTxos, pack);
            matRes = result.matLnk;
            newTxos = result.txos;
        }
        return { txos: newTxos, matLnk: matRes };
    }
    unpackLinkMatrix2(matLnk, txos, pack) {
        let newMatLnk = [];
        let newTxos = txos;
        if (matLnk.length > 0) {
            if (pack.packType === PackType.INPUTS) {
                const newInputs = new CustomMap([...txos.inputs]);
                const idx = this.unpackTxos(txos.inputs, pack, newInputs);
                newTxos = {
                    inputs: new CustomMap([...newInputs]),
                    outputs: new CustomMap([...txos.outputs]),
                };
                const nbIns = txos.inputs.size + pack.ins.length - 1;
                const nbOuts = txos.outputs.size;
                const newMatLnkFinal = Array.from({ length: nbOuts }, () => []);
                for (let i = 0; i < nbOuts; i++) {
                    const line = Array.from({ length: nbIns }).fill(0);
                    for (let j = 0; j < nbIns; j++) {
                        if (j < idx) {
                            line[j] = matLnk[i][j];
                        }
                        else if (j >= idx + pack.ins.length) {
                            line[j] = matLnk[i][j - pack.ins.length + 1];
                        }
                        else {
                            line[j] = matLnk[i][idx];
                        }
                    }
                    newMatLnkFinal[i] = line;
                }
                newMatLnk = newMatLnkFinal;
            }
        }
        return { txos: newTxos, matLnk: newMatLnk };
    }
    unpackTxos(currentTxos, pack, unpackedTxos) {
        const txoKeys = [...currentTxos.keys()];
        let idx;
        for (idx = 0; idx < txoKeys.length; idx++) {
            if (pack.lbl === txoKeys[idx]) {
                break;
            }
        }
        unpackedTxos.clear();
        const keysIterator = currentTxos.keys();
        for (let i = 0; i < idx; i++) {
            const key = keysIterator.next().value;
            unpackedTxos.set(key, currentTxos.get(key));
        }
        for (const [key, val] of pack.ins) {
            unpackedTxos.set(key, val);
        }
        keysIterator.next();
        let next = keysIterator.next();
        while (!next.done) {
            unpackedTxos.set(next.value, currentTxos.get(next.value));
            next = keysIterator.next();
        }
        return idx;
    }
    prepareData(txos) {
        const allInAgg = this.prepareTxos(txos.inputs);
        const allOutAgg = this.prepareTxos(txos.outputs);
        return { inAgg: allInAgg, outAgg: allOutAgg };
    }
    prepareTxos(initialTxos) {
        const txos = new CustomMap();
        const sortedEntries = [...initialTxos.entries()].sort((a, b) => b[1] - a[1]);
        for (const [key, value] of sortedEntries) {
            if (value > 0) {
                txos.set(key, value);
            }
        }
        const allVal = [...txos.values()];
        const allIndexes = Array.from({ length: txos.size }, (_, i) => i);
        const nbAggregates = 2 ** allIndexes.length;
        const allAggIndexes = ListsUtils.powerSet(allIndexes, this.Logger);
        const allAggVal = [];
        const PROGRESS_ID = "prepareTxos";
        for (const array of allAggIndexes) {
            allAggVal.push(array.reduce((acc, val) => acc + allVal[val], 0));
            Utils.logProgress(this.Logger, PROGRESS_ID, allAggVal.length, nbAggregates);
        }
        Utils.logProgressDone(this.Logger, PROGRESS_ID, nbAggregates);
        return { txos, allAggIndexes, allAggVal };
    }
    checkLimitOk(txos) {
        const lenIn = txos.inputs.size;
        const lenOut = txos.outputs.size;
        const maxCard = Math.max(lenIn, lenOut);
        if (this.maxTxos !== null && maxCard > this.maxTxos) {
            this.Logger.logInfo("maxTxos limit reached!");
            return false;
        }
        return true;
    }
}
TxosLinker.MAX_DURATION = 180;
TxosLinker.MAX_NB_TXOS = 12;
TxosLinker.MARKER_FEES = "FEES";
TxosLinker.MARKER_PACK = "PACK_I";
//# sourceMappingURL=txos-linker.js.map