import { Big } from "big.js";
import { IntraFees } from "../linker/intra-fees.js";
import { TxosLinkerResult } from "../linker/txos-linker-result.js";
import { TxosLinker } from "../linker/txos-linker.js";
import { ListsUtils } from "../utils/lists-utils.js";
import { CustomMap } from "../utils/utils.js";
import { TxProcessorConst } from "./tx-processor-const.js";
import { TxProcessorResult } from "./tx-processor-result.js";
export class TxProcessor {
    constructor(maxDuration, maxTxos, logger) {
        this.maxDuration = maxDuration;
        this.maxTxos = maxTxos;
        this.Logger = logger;
    }
    processTx(txos, maxCjIntrafeesRatio, linkerOptions) {
        this.Logger.logDebug(`Processing·tx:·${txos.inputs.size}·inputs,·${txos.outputs.size}·outputs`);
        const filteredIns = this.filterTxos(txos.inputs, TxProcessorConst.MARKER_INPUT);
        const filteredOuts = this.filterTxos(txos.outputs, TxProcessorConst.MARKER_OUTPUT);
        const sumInputs = [...filteredIns.txos.values()].reduce((prev, curr) => prev + curr, 0);
        const sumOutputs = [...filteredOuts.txos.values()].reduce((prev, curr) => prev + curr, 0);
        const fees = sumInputs - sumOutputs;
        let intraFees = new IntraFees(0, 0);
        let result;
        if (filteredIns.txos.size <= 1 || filteredOuts.txos.size === 1) {
            result = new TxosLinkerResult(1, [], new Set(), {
                inputs: filteredIns.txos,
                outputs: filteredOuts.txos,
            });
        }
        else {
            const filteredTxos = {
                inputs: filteredIns.txos,
                outputs: filteredOuts.txos,
            };
            const linker = new TxosLinker(fees, this.maxDuration, this.maxTxos, this.Logger);
            let linkedIns = [];
            let linkedOuts = [];
            if (linkerOptions.mergeInputs) {
                linkedIns = this.getLinkedTxos(filteredIns);
            }
            if (linkerOptions.mergeOutputs) {
                linkedOuts = this.getLinkedTxos(filteredOuts);
            }
            if (maxCjIntrafeesRatio > 0) {
                const lsFilteredIns = [];
                for (const txoId of filteredIns.txos.keys()) {
                    const set = new Set();
                    set.add(txoId);
                    lsFilteredIns.push(set);
                }
                lsFilteredIns.push(...linkedIns);
                const insToMerge = [];
                insToMerge.push(...lsFilteredIns, ...linkedIns);
                const maxNbPtcpts = ListsUtils.mergeSets(insToMerge).length;
                const cjPattern = this.checkCoinjoinPattern(filteredOuts.txos, maxNbPtcpts);
                if (cjPattern != null) {
                    intraFees = this.computeCoinjoinIntrafees(cjPattern.nbPtcpts, cjPattern.cjAmount, maxCjIntrafeesRatio);
                }
            }
            const linkedTxos = [];
            linkedTxos.push(...linkedIns, ...linkedOuts);
            result = linker.process(filteredTxos, linkedTxos, linkerOptions, intraFees);
        }
        const nbTxosPrfctCj = this.getClosestPerfectCoinjoin(filteredIns.txos.size, filteredOuts.txos.size);
        const nbCmbnPrfctCj = this.computeCmbnsPerfectCj(nbTxosPrfctCj.nbIns, nbTxosPrfctCj.nbOuts);
        let efficiency = null;
        if (nbCmbnPrfctCj != null) {
            efficiency = this.computeWalletEfficiency(result.nbCmbn, nbCmbnPrfctCj);
        }
        const txoIns = this.postProcessTxos(result.txos.inputs, filteredIns.mapIdAddr);
        const txoOuts = this.postProcessTxos(result.txos.outputs, filteredOuts.mapIdAddr);
        return new TxProcessorResult(result.nbCmbn, result.matLnkCombinations, result.computeMatLnkProbabilities(), result.computeEntropy(), result.dtrmLnksById, { inputs: txoIns, outputs: txoOuts }, fees, intraFees, efficiency, nbCmbnPrfctCj, nbTxosPrfctCj);
    }
    getLinkedTxos(filteredTxos) {
        const linkedTxos = [];
        for (const id of filteredTxos.txos.keys()) {
            let setIns = new Set();
            setIns.add(id);
            let setAddr = new Set();
            setAddr.add(filteredTxos.mapIdAddr.get(id));
            for (let i = 0; i < linkedTxos.length; i++) {
                const [k, v] = linkedTxos[i];
                if ([...k].some((addr) => setAddr.has(addr))) {
                    setIns = new Set([...setIns, ...v]);
                    setAddr = new Set([...setAddr, ...k]);
                    linkedTxos.splice(i, 1);
                    i--;
                }
            }
            linkedTxos.push([setAddr, setIns]);
        }
        const result = [];
        for (const sets of linkedTxos) {
            if (sets[1].size > 1) {
                result.push(sets[1]);
            }
        }
        return result;
    }
    filterTxos(txos, prefix) {
        const filteredTxos = new CustomMap();
        const mapIdAddr = new Map();
        for (const entry of txos.entries()) {
            if (entry[1] > 0) {
                const txoId = prefix + mapIdAddr.size;
                filteredTxos.set(txoId, entry[1]);
                mapIdAddr.set(txoId, entry[0]);
            }
        }
        return { txos: filteredTxos, mapIdAddr };
    }
    postProcessTxos(txos, mapIdAddr) {
        const results = new CustomMap();
        for (const entry of txos.entries()) {
            if (entry[0].startsWith(TxProcessorConst.MARKER_INPUT) ||
                entry[0].startsWith(TxProcessorConst.MARKER_OUTPUT)) {
                results.set(mapIdAddr.get(entry[0]), entry[1]);
            }
            else {
                results.set(entry[0], entry[1]);
            }
        }
        return results;
    }
    checkCoinjoinPattern(txoOuts, maxNbEntities) {
        if (maxNbEntities < 2) {
            return null;
        }
        const nbOutsByAmount = new Map();
        for (const entry of txoOuts.entries()) {
            const amont = entry[1];
            let nb = nbOutsByAmount.get(amont) ?? 0;
            nb++;
            nbOutsByAmount.set(amont, nb);
        }
        const nbTxoOuts = txoOuts.size;
        let isCj = false;
        let resNbPtcpts = 0;
        let resAmount = 0;
        for (const entry of nbOutsByAmount.entries()) {
            const amount = entry[0];
            const nbOutsForAmount = entry[1];
            if (nbOutsForAmount > 1) {
                const maxNbPtcpts = Math.min(nbOutsForAmount, maxNbEntities);
                const condTxoOuts = nbTxoOuts <= 2 * maxNbPtcpts;
                const condMaxPtcpts = maxNbPtcpts >= resNbPtcpts;
                const condMaxAmount = amount > resAmount;
                if (condTxoOuts && condMaxPtcpts && condMaxAmount) {
                    isCj = true;
                    resNbPtcpts = maxNbPtcpts;
                    resAmount = amount;
                }
            }
        }
        if (!isCj) {
            return null;
        }
        return { nbPtcpts: resNbPtcpts, cjAmount: resAmount };
    }
    computeCoinjoinIntrafees(nbPtcpts, cjAmount, prctMax) {
        const feeMaker = Math.round(cjAmount * prctMax);
        const feeTaker = feeMaker * (nbPtcpts - 1);
        return new IntraFees(feeMaker, feeTaker);
    }
    computeWalletEfficiency(nbCmbn, nbCmbnPrfctCj) {
        if (nbCmbn === 1) {
            return 0;
        }
        return new Big(nbCmbn).div(new Big(nbCmbnPrfctCj.toString())).toNumber();
    }
    getClosestPerfectCoinjoin(nbIns, nbOuts) {
        if (nbIns > nbOuts) {
            const nbInsInitial = nbIns;
            nbIns = nbOuts;
            nbOuts = nbInsInitial;
        }
        if (nbOuts % nbIns === 0) {
            return { nbIns, nbOuts };
        }
        const tgtRatio = Math.floor(1 + nbOuts / nbIns);
        return { nbIns, nbOuts: nbIns * tgtRatio };
    }
    computeCmbnsPerfectCj(nbIns, nbOuts) {
        if (nbIns > nbOuts) {
            const nbInsInitial = nbIns;
            nbIns = nbOuts;
            nbOuts = nbInsInitial;
        }
        if (nbOuts % nbIns !== 0) {
            return null;
        }
        if (nbIns <= 1 || nbOuts <= 1) {
            return 1n;
        }
        else if (nbIns <= 20 && nbOuts <= 60) {
            return TxProcessorConst.getNbCmbnPrfctCj(nbIns, nbOuts) || null;
        }
        return null;
    }
}
//# sourceMappingURL=tx-processor.js.map