var _a;
import { LogLevel } from "../beans/boltzmann-settings.js";
function isNodeJS() {
    return (typeof process !== "undefined" &&
        process.versions != null &&
        process.versions.node != null &&
        typeof window === "undefined");
}
const v8 = isNodeJS() ? await import("node:v8") : null;
export const Logger = (logLevel) => {
    return {
        logLevel,
        logInfo: (...msg) => {
            logLevel >= LogLevel.INFO && console.log(...msg);
        },
        logDebug: (...msg) => {
            logLevel >= LogLevel.DEBUG && console.log(...msg);
        },
    };
};
export class Utils {
    static logMemory(logger, msg) {
        if (v8) {
            const totalHeapSize = v8.getHeapStatistics().total_available_size / Utils.BYTE_TO_MB;
            const usedHeapSize = v8.getHeapStatistics().used_heap_size / Utils.BYTE_TO_MB;
            logger.logInfo(`${totalHeapSize - usedHeapSize}M free - ${msg ?? ""}`);
            if (usedHeapSize > Utils.maxMemUsed) {
                Utils.maxMemUsed = usedHeapSize;
            }
        }
        else {
            logger.logInfo(msg);
        }
    }
    static logProgress(logger, progressId, current, target, msg = "") {
        let progress = Utils.progressLast.get(progressId);
        const now = Date.now();
        if (progress && now - progress.last < Utils.LOG_PROGRESS_FREQUENCY * 1000) {
            return;
        }
        if (progress) {
            progress.update(current, target);
        }
        else {
            progress = new Progress(progressId, current, target);
            Utils.progressLast.set(progressId, progress);
        }
        Utils.logMemory(logger, `${progress.getProgress()} ${msg}`);
    }
    static logProgressDone(logger, progressId, target, msg = "") {
        const progress = Utils.progressLast.get(progressId);
        if (progress) {
            Utils.progressResult.push(progress);
            Utils.progressLast.delete(progressId);
        }
        const str = progress
            ? progress.done(target, msg)
            : `[${progressId}] (no iteration) ${msg}`;
        Utils.logMemory(logger, str);
    }
    static getProgressResult() {
        return Utils.progressResult;
    }
    static getMaxMemUsed() {
        return Utils.maxMemUsed;
    }
    static duration(seconds) {
        let minutes = 0;
        let hours = 0;
        let days = 0;
        if (seconds >= 60) {
            minutes = Math.floor(seconds / 60);
            seconds %= 60;
            if (minutes >= 60) {
                hours = Math.floor(minutes / 60);
                minutes %= 60;
            }
            if (hours >= 24) {
                days = Math.floor(hours / 24);
                hours %= 24;
            }
        }
        let sb = "";
        if (days > 0) {
            sb += `${days}d`;
        }
        if (hours > 0) {
            sb += `${hours}h`;
        }
        if (minutes > 0) {
            sb += `${minutes}m`;
        }
        if (seconds > 0 || sb.length === 0) {
            sb += `${seconds}s`;
        }
        return sb;
    }
}
Utils.BYTE_TO_MB = 1024 * 1024;
Utils.LOG_PROGRESS_FREQUENCY = 30;
Utils.progressLast = new Map();
Utils.progressResult = [];
Utils.maxMemUsed = 0;
class Progress {
    constructor(name, current, target) {
        this.name = name;
        this.start = Date.now();
        this.last = this.start;
        this.current = current === 0 ? 1 : current;
        this.target = target === 0 ? 1 : target;
        this.rate = 0;
        this.msg = "";
    }
    computeElapsed() {
        return this.last - this.start;
    }
    getProgress() {
        const elapsed = this.computeElapsed();
        const todo = this.target - this.current + 1;
        const donePercent = Math.floor((this.current * 100) / this.target);
        let str = `[${this.name}] ${donePercent}% ${this.current}/${this.target}`;
        if (elapsed > 0 || this.rate > 0) {
            const eta = this.rate > 0 ? Math.floor(todo / this.rate) : 0;
            str += ` since ${Utils.duration(elapsed / 1000)}, ${this.rate.toFixed(3)}/s, ETA ${Utils.duration(eta)}...`;
        }
        return str;
    }
    getResult() {
        const elapsed = this.computeElapsed();
        return `[${this.name}] ${this.target}x in ${Utils.duration(elapsed)}s (${this.rate.toFixed(3)}/s) ${this.msg}`;
    }
    update(current, target) {
        const now = Date.now();
        const newRate = current / ((now - this.start) / 1000);
        this.rate = this.rate === 0 ? newRate : (this.rate + newRate) / 2;
        this.current = current;
        this.last = now;
        this.target = target;
    }
    done(target, msg) {
        this.last = Date.now();
        this.current = target;
        this.target = target;
        this.msg = msg;
        if (this.rate === 0) {
            this.rate = target;
        }
        const elapsed = this.computeElapsed();
        return `[${this.name}] 100% ${target} done in ${elapsed / 1000}s ${msg}`;
    }
}
export class CustomMap {
    constructor(entries) {
        this[_a] = "CustomMap";
        this.state = entries ? [...entries] : [];
    }
    [Symbol.iterator]() {
        return this.state[Symbol.iterator]();
    }
    entries() {
        return this.state[Symbol.iterator]();
    }
    keys() {
        return this.state.map(([k, _]) => k)[Symbol.iterator]();
    }
    values() {
        return this.state.map(([_, v]) => v)[Symbol.iterator]();
    }
    clear() {
        this.state = [];
    }
    delete(key) {
        const index = this.state.findIndex(([k, _]) => k === key);
        if (index !== -1) {
            this.state.splice(index, 1);
            return true;
        }
        return false;
    }
    forEach(callbackfn, thisArg) {
        this.state.forEach(([k, v]) => callbackfn.call(thisArg, v, k, this));
    }
    get(key) {
        const entry = this.state.find(([k, _]) => k === key);
        return entry ? entry[1] : undefined;
    }
    has(key) {
        return this.state.some(([k, _]) => k === key);
    }
    set(key, value) {
        this.state.push([key, value]);
        return this;
    }
    get size() {
        return this.state.length;
    }
}
_a = Symbol.toStringTag;
//# sourceMappingURL=utils.js.map